# FIAIF is an Intelligent firewall, version: $Revision: 1.88 $
#
# description: Automates a packet filtering firewall with iptables.
#
# Script Author:	Anders Fugmann <afu at fugmann dot dhs dot org>
# 
# FIAIF is an Intelligent firewall
# Copyright (C) 2002-2003 Anders Peter Fugmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Please drop me a line and give me your thoughts on this software.
# Patches are more than welcome.

###############################################################################
# Iptables. Log errornous lines
###############################################################################
function IPTABLES ()
{
    local ERR=0
    if [ "${TEST}" = "1" ]; then
	echo "iptables $@" >> ${TEST_FILE}
    else
        iptables $@ || ERR=1
    fi

    if [ "${ERR}" -eq 1 ]; then
        echo "Error: iptables $@" 1>&2
        let IPTABLES_ERRORS++
    fi
}

###############################################################################
# Call tc
###############################################################################
function TC ()
{
    local ERR=0
    if (( TEST == 1 )); then
	echo "tc $@" >> ${TEST_FILE}
    else
        tc $@ || ERR=1
    fi

    if (( ERR == 1 )); then
        echo "Error: tc $@" 1>&2
    fi
}

###############################################################################
# debug_out 
# write textual messages to file if testing. 
###############################################################################
function debug_out ()
{
    if (( TEST == 1 )); then
	echo "### $@" >> ${TEST_FILE}
    fi
}

###############################################################################
# print_err 
# write textual messages to strdout and file if testing. 
###############################################################################
function print_err ()
{
    echo "### $@" 
    if (( TEST == 1 )); then
	echo "### $@" >> ${TEST_FILE}
    fi
}



###############################################################################
# Print copyright and the version number
###############################################################################
function print_version ()
{
    local VERSION=$(<${VERSION_FILE})
    echo "FIAIF ver. $VERSION, by Anders Fugmann (C) 2002-2003"
    #debug_out "FIAIF ver. $VERSION, by Anders Fugmann (C) 2002-2003"     
}

###############################################################################
# get_protocol_port_ip
# param: 
#   <IN|OUT> [protocol [port[,port]* [ip[/mask]] ] ]
#
# returns: RETURN
###############################################################################
function get_protocol_port_ip ()
{
    local DIRECTION=$1
    local PROTOCOL=$2
    shift 2
	
    RETURN="-p ${PROTOCOL}"
    local PORTS=""
    case ${PROTOCOL} in
	TCP | tcp | UDP | udp)
	    PORTS=$1
	    shift
	    # Test to see if the port is a range
	    if [[ ! ${PORTS} == @(ALL|all) ]]; then
		if [[ ${PORTS} == *,* ]]; then 
		    RETURN="$RETURN -m multiport --dports ${PORTS}"
		else
		    RETURN="$RETURN --dport ${PORTS}"
		fi
	    fi
	    ;;
	ICMP|icmp)
	    PORTS=$1
	    shift
	    if [[ ! ${PORTS} == @(ALL|all) ]]; then
		RETURN="$RETURN --icmp-type ${PORTS}"
	    fi
	    ;;
    esac
    # IP: xx/yy=>xx/yy
    local IP=$1
    shift

    if [[ -n "${IP}" ]]; then
	local SRC_IP=${IP%*=>*}
	local DST_IP=${IP#*=>*}
	# Maintain backward compability with old configuration syntax
	if [ "${SRC_IP}=>${DST_IP}" != ${IP} ]; then
	    
	    print_err "Old style IP's should not be used."
	    print_err "Use xxx.xxx.xxx.xxx/yy=>zzz.zzz.zzz.zzz/vv style instead."
	    let RULE_ERRORS++

	    if [ "${IP}" != "0.0.0.0/0" ]; then
		case ${DIRECTION} in
		    IN)
			RETURN="$RETURN --src ${IP}"
			;;
		    OUT)
			RETURN="$RETURN --dst ${IP}"
			;;
		esac
	    fi
	else
	    if [ "${SRC_IP}" != */0 ]; then
		RETURN="$RETURN --src ${SRC_IP} "
	    fi
	    if [ "${DST_IP}" != */0 ]; then
		RETURN="$RETURN --dst ${DST_IP}"
	    fi
	fi
    else
	print_err ""
	print_err "Error in rules."
	print_err "All INPUT, OUTPUT, FORWARD, MARK and LIMIT rules"
	print_err "should specify src and destination".
	let RULE_ERRORS++
    fi
}


###############################################################################
# Give staus of the firewall
###############################################################################
function iptables_status ()
{
	IPTABLES -nvL
	IPTABLES -t nat -nvL
	IPTABLES -t mangle -nvL
}

###############################################################################
# Restore the firewall rules as saved with save_rules
###############################################################################
function restore_rules ()
{
    local STATE_FILE=$1
    echo -n "Restoring rules: "
    if [[ -f ${STATE_FILE} ]]; then
	iptables-restore -c < ${STATE_FILE}
    else
	echo -n "File '${STATE_FILE}' not found " 
    fi
    echo "Done."
}

###############################################################################
# Give staus of the firewall
###############################################################################
function save_rules ()
{
    local STATE_FILE=$1
    echo -n "Saving rules: "
    # Make sure that this file is not tampered.
    rm -f ${STATE_FILE}
    touch ${STATE_FILE}
    chown root:root ${STATE_FILE}
    chmod 600 ${STATE_FILE}
    iptables-save -c >> ${STATE_FILE}
    echo "Done."
    
}

###############################################################################
# apply_script: 
###############################################################################
function apply_script
{
    local SCRIPT=$1
    local LENGTH=$2
    local VAR
    local I
    for ((I=0;I<${LENGTH};I++)); do
	VAR=${SCRIPT}[$I]
	debug_out "${VAR}=${!VAR}"
        if (( TEST == 1 )); then
	    echo "${!VAR}" >> ${TEST_FILE}
	else
	    ${!VAR}
	fi
    done
}

###############################################################################
# Get the name of a service based on port and protocol
# RESULT holds the return value.
# The result is cached.
###############################################################################
function get_service_name
{
    local PROTOCOL=$1
    local PORT=$2
    #Lookup in the cache
    RESULT=""
    local I
    if [[ -n "${PORT}" ]]; then
	for ((I=0;I<${#SERVICE_CACHE[*]};I++));do
	    if [[ "${SERVICE_CACHE[I]%:*}" = "${PORT}/${PROTOCOL}" ]];then
		RESULT=${SERVICE_CACHE[I]#*:}
		return
	    fi
	done
	
	RESULT=$(grep -i -e "[^0-9]${PORT}/${PROTOCOL}" ${SERVICES} | cut -f1)
	if [[ -z "${RESULT}" ]]; then
	    RESULT="${PORT}"
	fi
	SERVICE_CACHE[${#SERVICE_CACHE[*]}]="${PORT}/${PROTOCOL}:$RESULT"
    else
	RESULT=""
    fi
}

###############################################################################
# Get the name of a ip-address. 
# RESULT holds the return value.
# Result is pooled for later use.
###############################################################################
function get_host_name ()
{
    local IP=$1
    RESULT=${IP}
    if [[ -n "${IP}" ]]; then
	local I
	for ((I=0;I<${#HOST_CACHE[*]};I++));do
	    if [[ "${HOST_CACHE[I]}" = ${IP}:* ]];then
		RESULT=${HOST_CACHE[I]#*:}
		return
	    fi
	done
	RESULT=$(dig +short +tries=1 +noauthority -x ${IP})
	if [[ -n "${RESULT}" ]]; then
	    IP=${RESULT%%.}
	fi
	# Always put result into the cache to save time looking up again.
	HOST_CACHE[${#HOST_CACHE[*]}]="${IP}:$RESULT"
    fi
}

###############################################################################
# Load modules. 
###############################################################################
function load_modules ()
{
    local MODULE
    for MODULE in ${MODULES}; do
	if (( TEST == 1 )); then
	    echo "modprobe ${MODULE}" >> ${TEST_FILE}
	else
	    modprobe ${MODULE}
	fi
    done
}

###############################################################################
# Unload modules. 
###############################################################################
function unload_modules ()
{
    local MODULE
    for MODULE in ${MODULES}; do
	if (( TEST == 1 )); then
	    echo "rmmod ${MODULE}" >> ${TEST_FILE}
	else
	    rmmod ${MODULE} > /dev/null 2>&1
	fi  
    done
}

###############################################################################
# Convert a mask to a number. 
###############################################################################
function mask_to_number ()
{
    local MASK=$1
    RESULT=1
    case ${MASK} in
	1)
	    RESULT=128
	    ;;
	2)
	    RESULT=64
	    ;;
	3)
	    RESULT=32
	    ;;
	4)
	    RESULT=16
	    ;;
	5)
	    RESULT=8
	    ;;
	6)
	    RESULT=4
	    ;;
	7)
	    RESULT=2
	    ;;
	8)
	    RESULT=1
	    ;;
    esac
}
###############################################################################
# Convert a mask to a number. 
###############################################################################
function bitmap_to_mask ()
{
    local MASK=$1
    RESULT=0
    local BITS
    local i
    for ((i=0;i<4;i++)); do
	
	case ${MASK%%.*} in
	    255)
		BITS=8
		;;
	    254)
		BITS=7
		;;
	    252)
		BITS=6
		;;
	    248)
		BITS=5
		;;
	    240)
		BITS=4
		;;
	    224)
		BITS=3
		;;
	    192)
		BITS=2
		;;
	    128)
		BITS=1
		;;
	    0)
		BITS=0
	    ;;
	esac
	
	RESULT=$(( ${RESULT} + ${BITS} ))
	MASK=${MASK#*.}
    done
}

###############################################################################
# Test if an IP is in the given network
# Param: network mask ip
# Where mask is a single number.
###############################################################################
function ip_in_network ()
{
    local NET=$1
    local MASK=$2
    local IP=$3

    local NUMBER
    # If mask is in dotted notation, convert it to bitmask
    if [[ "${MASK}" == +(+([[:digit:]]).)+([[:digit:]]) ]]; then
	bitmap_to_mask ${MASK}
	MASK=${RESULT}
    fi

    while (( ${MASK} >= 1 )); do
	mask_to_number ${MASK}
	NUMBER=${RESULT}
	
	NET_FIELD=$(( ${NET%%.*} / ${NUMBER} ))
	IP_FIELD=$(( ${IP%%.*} / ${NUMBER} ))

	#echo ${NET_FIELD} -ne ${IP_FIELD}
	if (( NET_FIELD != IP_FIELD )); then
	    return -1
	    break
	fi

	NET=${NET#*.}
	IP=${IP#*.}
	MASK=$((${MASK} - 8))
    done
    return 0
}

###############################################################################
# Compare two ip numbers.
# Result: -1 ip1 lower than ip2
#          0 equal
#          1 ip1 greater than ip2
###############################################################################
function compare_ip ()
{
    RESULT=0
    
    declare -a IP1="( $(echo $1 | sed s/\\\./\ /g) )"
    declare -a IP2="( $(echo $2 | sed s/\\\./\ /g) )"
    local I
    for ((I=0;I<4;I++)); do
	if (( ${IP1[I]} > ${IP2[I]} ));then
	    RESULT=1
	    return
	elif (( ${IP1[I]} < ${IP2[I]} )); then
	    RESULT=-1
	    return
	fi
    done
}

###############################################################################
# Check that a zone exists
# Returns: RETURN=0|1
###############################################################################
function check_zone ()
{
    local ZONE=$1
    local ERRORMSG=$2
    RESULT=0
    local AZONE
    
    for AZONE in ${ZONES} ALL; do
	if [[ "${ZONE}" == "${AZONE}" ]]; then
	    RESULT=1
	    break
	fi
    done
    if [[ -n "${ERRORMSG}" ]] && (( ${RESULT} == 0 )); then
	local ZONEFILE 
	ZONEFILE=CONF_${NAME}
	ZONEFILE=${!ZONEFILE}
	echo -e "\n${ZONEFILE}: ${ERRORMSG}\n"
	debug_out "${ZONEFILE}: ${ERRORMSG}"
    fi
}

###############################################################################
# Locate zone based on ip and interface. All zones needs to be loaded.
# Param: <interface> <ip>
# Returns: Name of the zone
###############################################################################
function get_zone_name ()
{
    local ZONE_DEV=$1
    local ZONE_IP=$2

    RESULT="UNKNOWN"

    for ZONE in ${ZONES}; do
	GLOBAL=${ZONE}_GLOBAL; GLOBAL=${!GLOBAL}
	DYNAMIC=${ZONE}_DYNAMIC; DYNAMIC=${!DYNAMIC}
	NETS=${ZONE}_NETS; NETS=${!NETS}

	if [[ ${ZONE_DEV} == ${DEV} ]]; then
	    if (( GLOBAL == 1 || DYNAMIC == 1 )); then
		RESULT=${ZONE}
	    else
		for NET in ${NETS}; do
		    if ip_in_network ${NET%/*} ${NET#*/} ${ZONE_IP}; then
			RESULT=${ZONE}
		    fi
		done
	    fi
	fi
	if [[ "${RESULT}" != "UNKNOWN" ]]; then
	    break
	fi
    done
}
